@@ -39,6 +39,18 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #include <pthread.h>
 #include <signal.h>
 
+#ifdef HAVE_ENDIAN_H
+# include <endian.h>
+#else
+# define __LITTLE_ENDIAN	1234
+# define __BIG_ENDIAN		4321
+# if defined(__hpux)
+#   define __BYTE_ORDER __LITTLE_ENDIAN
+# else
+#   error Host has unknown byte-order.
+# endif
+#endif
+
 #ifdef __GNUC__
 # if (__GNUC__ > 3) || (__GNUC__ == 3 && __GNUC_MINOR__ > 2)
 #  define HIDDEN	__attribute__((visibility ("hidden")))
@@ -83,6 +95,7 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
    unconditionally and if -lpthread is around, they'll call the
    corresponding routines otherwise, they do nothing.  */
 
+#pragma weak pthread_mutex_init
 #pragma weak pthread_mutex_lock
 #pragma weak pthread_mutex_unlock
 
@@ -93,8 +106,17 @@ WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.  */
 #ifdef HAVE_IA64INTRIN_H
 # define HAVE_CMPXCHG
 # include <ia64intrin.h>
+  /*
+   * ecc v7.0 is broken: it's missing __sync_val_compare_and_swap()
+   * even though the ia64 ABI requires it.  Work around it:
+   */
+# ifndef __sync_val_compare_and_swap
+#  define __sync_val_compare_and_swap(x,y,z)		\
+	_InterlockedCompareExchange64_rel(x,y,z)
+# endif
+
 # define cmpxchg_ptr(_ptr,_o,_n)					\
-	((void *) __sync_val_compare_and_swap((long *) (_ptr),		\
+	((void *) __sync_val_compare_and_swap((volatile long *) (_ptr),	\
 					      (long) (_o), (long) (_n)))
 #endif
 
