@@ -210,7 +210,7 @@ extern void UNW_ARCH_OBJ(flush_cache)(unw_addr_space_t as,
 				      unw_word_t lo, unw_word_t hi);
 extern int UNW_ARCH_OBJ(set_caching_policy)(unw_addr_space_t as,
 					    unw_caching_policy_t policy);
-extern const char *UNW_ARCH_OBJ(regname) (int regnum);
+extern const char *UNW_ARCH_OBJ(regname) (unw_regnum_t regnum);
 
 extern int UNW_OBJ(init_local) (unw_cursor_t *c, unw_context_t *u);
 extern int UNW_OBJ(init_remote) (unw_cursor_t *c, unw_addr_space_t as,
