@@ -42,7 +42,8 @@ _UPT_access_reg (unw_addr_space_t as, unw_regnum_t reg, unw_word_t *val,
 
 #if DEBUG
   if (write)
-    debug (100, "%s: %s <- %lx\n", __FUNCTION__, unw_regname (reg), *val);
+    debug (100, "%s: %s <- %lx\n",
+	   __FUNCTION__, unw_regname (reg), (long) *val);
 #endif
 
 #if UNW_TARGET_IA64
@@ -236,10 +237,13 @@ _UPT_access_reg (unw_addr_space_t as, unw_regnum_t reg, unw_word_t *val,
     goto badreg;
 #endif
 
+#ifdef UNW_TARGET_IA64
  out:
+#endif
 #if DEBUG
   if (!write)
-    debug (100, "%s: %s -> %lx\n", __FUNCTION__, unw_regname (reg), *val);
+    debug (100, "%s: %s -> %lx\n",
+	   __FUNCTION__, unw_regname (reg), (long) *val);
 #endif
   return 0;
 
