@@ -37,8 +37,8 @@ static int create_state_record_for (struct cursor *c,
 
 typedef unsigned long unw_word;
 
-#define alloc_reg_state()	(mempool_alloc (&unw.state_record_pool))
-#define free_reg_state(rs)	(mempool_free (&unw.state_record_pool, rs))
+#define alloc_reg_state()	(mempool_alloc (&unw.reg_state_pool))
+#define free_reg_state(rs)	(mempool_free (&unw.reg_state_pool, rs))
 #define alloc_labeled_state()	(mempool_alloc (&unw.labeled_state_pool))
 #define free_labeled_state(s)	(mempool_free (&unw.labeled_state_pool, s))
 
@@ -134,7 +134,7 @@ decode_abreg (unsigned char abreg, int memory)
     case 0x62:
       return memory ? IA64_REG_PRI_UNAT_MEM : IA64_REG_PRI_UNAT_GR;
     case 0x63:
-      return IA64_REG_RP;
+      return IA64_REG_IP;
     case 0x64:
       return IA64_REG_BSP;
     case 0x65:
@@ -318,12 +318,7 @@ static inline void
 desc_abi (unsigned char abi, unsigned char context,
 	  struct ia64_state_record *sr)
 {
-  if ((abi == 0 && context == 's')	/* Linux signal trampoline */
-      || (abi == 1 && context == 1))	/* HP-UX signal trampoline */
-    sr->is_signal_frame = 1;
-  else
-    dprintf ("libunwind: ignoring unwabi(abi=0x%x,context=0x%x)\n",
-	     abi, context);
+  sr->abi_marker = (abi << 8) | context;
 }
 
 static inline void
@@ -559,9 +554,9 @@ desc_is_active (unsigned char qp, unw_word t, struct ia64_state_record *sr)
     return 0;
   if (qp > 0)
     {
-      if ((sr->pr_val & (1UL << qp)) == 0)
+      if ((sr->pr_val & ((unw_word_t) 1 << qp)) == 0)
 	return 0;
-      sr->pr_mask |= (1UL << qp);
+      sr->pr_mask |= ((unw_word_t) 1 << qp);
     }
   return 1;
 }
@@ -645,7 +640,7 @@ desc_spill_sprel_p (unsigned char qp, unw_word t, unsigned char abreg,
 #define UNW_REG_PR		IA64_REG_PR
 #define UNW_REG_RNAT		IA64_REG_RNAT
 #define UNW_REG_PSP		IA64_REG_PSP
-#define UNW_REG_RP		IA64_REG_RP
+#define UNW_REG_RP		IA64_REG_IP
 #define UNW_REG_UNAT		IA64_REG_UNAT
 
 /* Region headers.  */
@@ -715,7 +710,7 @@ lookup_preg (int regnum, int memory, struct ia64_state_record *sr)
     case UNW_IA64_AR_PFS:		preg = IA64_REG_PFS; break;
     case UNW_IA64_AR_RNAT:		preg = IA64_REG_RNAT; break;
     case UNW_IA64_AR_UNAT:		preg = IA64_REG_UNAT; break;
-    case UNW_IA64_BR + 0:		preg = IA64_REG_RP; break;
+    case UNW_IA64_BR + 0:		preg = IA64_REG_IP; break;
     case UNW_IA64_PR:			preg = IA64_REG_PR; break;
     case UNW_IA64_SP:			preg = IA64_REG_PSP; break;
 
@@ -743,8 +738,7 @@ lookup_preg (int regnum, int memory, struct ia64_state_record *sr)
       break;
 
     default:
-      dprintf ("unwind.%s: invalid register number %d\n",
-	       __FUNCTION__, regnum);
+      dprintf ("%s: invalid register number %d\n", __FUNCTION__, regnum);
       return NULL;
     }
   return sr->curr.reg + preg;
@@ -761,8 +755,8 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
   enum ia64_where where;
   unw_dyn_op_t *op;
   int32_t when, len;
+  int memory, ret;
   int8_t qp;
-  int memory;
 
   for (r = proc->regions; r; r = r->next)
     {
@@ -809,7 +803,7 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 		where = IA64_WHERE_BR;
 	      else
 		{
-		  dprintf ("unwind.%s: can't save to register number %d\n",
+		  dprintf ("%s: can't save to register number %d\n",
 			   __FUNCTION__, (int) op->reg);
 		  return -UNW_EBADREG;
 		}
@@ -839,7 +833,7 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 		{
 		  if (val & 0xf)
 		    {
-		      dprintf ("unwind.%s: frame-size %ld not an integer "
+		      dprintf ("%s: frame-size %ld not an integer "
 			       "multiple of 16\n",
 			       __FUNCTION__, (long) op->val);
 		      return -UNW_EINVAL;
@@ -848,7 +842,7 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 		}
 	      else
 		{
-		  dprintf ("unwind.%s: can only ADD to stack-pointer\n",
+		  dprintf ("%s: can only ADD to stack-pointer\n",
 			   __FUNCTION__);
 		  return -UNW_EBADREG;
 		}
@@ -872,6 +866,9 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 		pop (sr);
 	      new_ip = op->val + ((sr->when_target / 3) * 16
 				  + (sr->when_target % 3));
+	      ret = ia64_fetch_proc_info (c, new_ip, 1);
+	      if (ret < 0)
+		return ret;
 	      return create_state_record_for (c, sr, new_ip);
 
 	    case UNW_DYN_STOP:
@@ -885,8 +882,8 @@ parse_dynamic (struct cursor *c, struct ia64_state_record *sr)
 }
 
 
-static int
-get_proc_info (struct cursor *c, unw_word_t ip, int need_unwind_info)
+HIDDEN int
+ia64_fetch_proc_info (struct cursor *c, unw_word_t ip, int need_unwind_info)
 {
   int ret, dynamic = 1;
 
@@ -928,6 +925,8 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
   uint8_t *dp, *desc_end;
   int ret;
 
+  assert (c->pi_valid);
+
   /* build state record */
   memset (sr, 0, sizeof (*sr));
   for (r = sr->curr.reg; r < sr->curr.reg + IA64_NUM_PREGS; ++r)
@@ -936,23 +935,20 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
   sr->first_region = 1;
   sr->return_link_reg = 0;
 
-  ret = get_proc_info (c, ip, 1);
-  if (ret < 0)
-    return ret;
-
   if (!c->pi.unwind_info)
     {
       /* No info, return default unwinder (leaf proc, no mem stack, no
          saved regs), rp in b0, pfs in ar.pfs.  */
       debug (1, "unwind.parser: no unwind info for ip=0x%lx (gp=%lx)\n",
 	     (long) ip, (long) c->pi.gp);
-      sr->curr.reg[IA64_REG_RP].where = IA64_WHERE_BR;
-      sr->curr.reg[IA64_REG_RP].when = -1;
-      sr->curr.reg[IA64_REG_RP].val = 0;
+      sr->curr.reg[IA64_REG_IP].where = IA64_WHERE_BR;
+      sr->curr.reg[IA64_REG_IP].when = -1;
+      sr->curr.reg[IA64_REG_IP].val = 0;
       goto out;
     }
 
-  sr->when_target = (3 * ((ip & ~0xfUL) - c->pi.start_ip) / 16 + (ip & 0xfUL));
+  sr->when_target = (3 * ((ip & ~(unw_word_t) 0xf) - c->pi.start_ip) / 16
+		     + (ip & 0xf));
 
   if (c->pi.format == UNW_INFO_FORMAT_TABLE)
     {
@@ -989,11 +985,11 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
 
   /* If RP did't get saved, generate entry for the return link
      register.  */
-  if (sr->curr.reg[IA64_REG_RP].when >= sr->when_target)
+  if (sr->curr.reg[IA64_REG_IP].when >= sr->when_target)
     {
-      sr->curr.reg[IA64_REG_RP].where = IA64_WHERE_BR;
-      sr->curr.reg[IA64_REG_RP].when = -1;
-      sr->curr.reg[IA64_REG_RP].val = sr->return_link_reg;
+      sr->curr.reg[IA64_REG_IP].where = IA64_WHERE_BR;
+      sr->curr.reg[IA64_REG_IP].when = -1;
+      sr->curr.reg[IA64_REG_IP].val = sr->return_link_reg;
     }
 
   if (sr->when_target > sr->curr.reg[IA64_REG_BSP].when
@@ -1049,6 +1045,8 @@ create_state_record_for (struct cursor *c, struct ia64_state_record *sr,
   return 0;
 }
 
+/* The proc-info must be valid for IP before this routine can be
+   called.  */
 HIDDEN int
 ia64_create_state_record (struct cursor *c, struct ia64_state_record *sr)
 {
@@ -1079,6 +1077,6 @@ ia64_make_proc_info (struct cursor *c)
   if (c->as->caching_policy == UNW_CACHE_NONE
       || ia64_get_cached_proc_info (c) < 0)
     /* Lookup it up the slow way... */
-    return get_proc_info (c, c->ip, 0);
+    return ia64_fetch_proc_info (c, c->ip, 0);
   return 0;
 }
