@@ -49,18 +49,26 @@ cache_match (struct ia64_script *script, unw_word_t ip, unw_word_t pr)
 }
 
 static inline void
-flush_script_cache (unw_addr_space_t as, struct ia64_script_cache *cache)
+flush_script_cache (struct ia64_script_cache *cache)
 {
   int i;
 
-  for (i = 0; i < IA64_UNW_CACHE_SIZE; ++i)
-    cache->buckets[i].ip = 0;
+  cache->lru_head = IA64_UNW_CACHE_SIZE - 1;
+  cache->lru_tail = 0;
 
-  cache->generation = as->cache_generation;
+  for (i = 0; i < IA64_UNW_CACHE_SIZE; ++i)
+    {
+      if (i > 0)
+	cache->buckets[i].lru_chain = (i - 1);
+      cache->buckets[i].coll_chain = -1;
+      cache->buckets[i].ip = 0;
+    }
+  for (i = 0; i<IA64_UNW_HASH_SIZE; ++i)
+    cache->hash[i] = -1;
 }
 
 static inline struct ia64_script_cache *
-get_script_cache (unw_addr_space_t as)
+get_script_cache (unw_addr_space_t as, sigset_t *saved_sigmaskp)
 {
   struct ia64_script_cache *cache = &as->global_cache;
 
@@ -69,12 +77,31 @@ get_script_cache (unw_addr_space_t as)
     cache = &ia64_per_thread_cache;
 #endif
 
-  if (as->cache_generation != cache->generation)
-    flush_script_cache (as, cache);
+  if (likely (as->caching_policy == UNW_CACHE_GLOBAL))
+    {
+      sigprocmask (SIG_SETMASK, &unwi_full_sigmask, saved_sigmaskp);
+      mutex_lock (&cache->lock);
+    }
 
+  if (as->cache_generation != cache->generation)
+    {
+      flush_script_cache (cache);
+      cache->generation = as->cache_generation;
+    }
   return cache;
 }
 
+static inline void
+put_script_cache (unw_addr_space_t as, struct ia64_script_cache *cache,
+		  sigset_t *saved_sigmaskp)
+{
+  if (likely (as->caching_policy == UNW_CACHE_GLOBAL))
+    {
+      mutex_unlock (&cache->lock);
+      sigprocmask (SIG_SETMASK, saved_sigmaskp, NULL);
+    }
+}
+
 static struct ia64_script *
 script_lookup (struct ia64_script_cache *cache, struct cursor *c)
 {
@@ -111,28 +138,17 @@ script_lookup (struct ia64_script_cache *cache, struct cursor *c)
 HIDDEN int
 ia64_get_cached_proc_info (struct cursor *c)
 {
-  int global_cache = (c->as->caching_policy == UNW_CACHE_GLOBAL);
   struct ia64_script_cache *cache;
   struct ia64_script *script;
   sigset_t saved_sigmask;
 
-  cache = get_script_cache (c->as);
-
-  if (likely (global_cache))
-    {
-      sigprocmask (SIG_SETMASK, &unwi_full_sigmask, &saved_sigmask);
-      mutex_lock (&cache->lock);
-    }
-
-  script = script_lookup (cache, c);
-  if (script)
-    c->pi = script->pi;
-
-  if (likely (global_cache))
-    {
-      mutex_unlock (&cache->lock);
-      sigprocmask (SIG_SETMASK, &saved_sigmask, NULL);
-    }
+  cache = get_script_cache (c->as, &saved_sigmask);
+  {
+    script = script_lookup (cache, c);
+    if (script)
+      c->pi = script->pi;
+  }
+  put_script_cache (c->as, cache, &saved_sigmask);
   return script ? 0 : -UNW_ENOINFO;
 }
 
@@ -184,6 +200,7 @@ script_new (struct ia64_script_cache *cache, unw_word_t ip)
   script->ip = ip;
   script->hint = 0;
   script->count = 0;
+  script->abi_marker = 0;
   return script;
 }
 
@@ -219,15 +236,15 @@ emit_nat_info (struct ia64_state_record *sr, int i, struct ia64_script *script)
   switch (r->where)
     {
     case IA64_WHERE_GR:
-      val = IA64_LOC (r->val, 0);
+      opc = IA64_INSN_SET_REG;
+      val = r->val;
       break;
 
     case IA64_WHERE_FR:
-      val = 0;			/* value doesn't matter... */
-      break;
+      break;				/* value doesn't matter... */
 
     case IA64_WHERE_BR:
-      val = IA64_LOC (0, 0);	/* no NaT bit */
+      /* val==0 results in IA64_LOC_NULL, i.e., no NaT bit */
       break;
 
     case IA64_WHERE_PSPREL:
@@ -241,7 +258,7 @@ emit_nat_info (struct ia64_state_record *sr, int i, struct ia64_script *script)
       return;
     }
   insn.opc = opc;
-  insn.dst = unw.preg_index[i];
+  insn.dst = i;
   insn.val = val;
   script_emit (script, insn);
 }
@@ -273,7 +290,7 @@ compile_reg (struct ia64_state_record *sr, int i, struct ia64_script *script)
 	}
       else if (rval >= 4 && rval <= 7)
 	/* register got spilled to a preserved register */
-	val = unw.preg_index[IA64_REG_R4 + (rval - 4)];
+	val = IA64_REG_R4 + (rval - 4);
       else
 	{
 	  /* register got spilled to a scratch register */
@@ -284,9 +301,9 @@ compile_reg (struct ia64_state_record *sr, int i, struct ia64_script *script)
 
     case IA64_WHERE_FR:
       if (rval <= 5)
-	val = unw.preg_index[IA64_REG_F2 + (rval - 1)];
+	val = IA64_REG_F2 + (rval - 1);
       else if (rval >= 16 && rval <= 31)
-	val = unw.preg_index[IA64_REG_F16 + (rval - 16)];
+	val = IA64_REG_F16 + (rval - 16);
       else
 	{
 	  opc = IA64_INSN_MOVE_SCRATCH;
@@ -296,7 +313,7 @@ compile_reg (struct ia64_state_record *sr, int i, struct ia64_script *script)
 
     case IA64_WHERE_BR:
       if (rval >= 1 && rval <= 5)
-	val = unw.preg_index[IA64_REG_B1 + (rval - 1)];
+	val = IA64_REG_B1 + (rval - 1);
       else
 	{
 	  opc = IA64_INSN_MOVE_SCRATCH;
@@ -322,7 +339,7 @@ compile_reg (struct ia64_state_record *sr, int i, struct ia64_script *script)
       break;
     }
   insn.opc = opc;
-  insn.dst = unw.preg_index[i];
+  insn.dst = i;
   insn.val = val;
   script_emit (script, insn);
   if (is_preserved_gr)
@@ -330,11 +347,10 @@ compile_reg (struct ia64_state_record *sr, int i, struct ia64_script *script)
 
   if (i == IA64_REG_PSP)
     {
-      /* info->psp must contain the _value_ of the previous sp, not
-	 it's save location.  We get this by dereferencing the value
-	 we just stored in info->psp: */
-      insn.opc = IA64_INSN_LOAD;
-      insn.dst = insn.val = unw.preg_index[IA64_REG_PSP];
+      /* c->psp must contain the _value_ of the previous sp, not it's
+	 save-location.  We get this by dereferencing the value we
+	 just stored in loc[IA64_REG_PSP]: */
+      insn.opc = IA64_INSN_LOAD_PSP;
       script_emit (script, insn);
     }
 }
@@ -360,8 +376,7 @@ build_script (struct cursor *c, struct ia64_script *script)
       && sr.curr.reg[IA64_REG_PSP].val != 0)
     {
       /* new psp is psp plus frame size */
-      insn.opc = IA64_INSN_ADD;
-      insn.dst = struct_offset (struct cursor, psp) / 8;
+      insn.opc = IA64_INSN_INC_PSP;
       insn.val = sr.curr.reg[IA64_REG_PSP].val;	/* frame size */
       script_emit (script, insn);
     }
@@ -387,14 +402,7 @@ build_script (struct cursor *c, struct ia64_script *script)
   for (i = IA64_REG_PSP; i < IA64_NUM_PREGS; ++i)
     compile_reg (&sr, i, script);
 
-  if (sr.is_signal_frame)
-    {
-      insn.opc = IA64_INSN_SET;
-      insn.dst = struct_offset (struct cursor, is_signal_frame) / 8;
-      insn.val = 1;
-      script_emit (script, insn);
-    }
-
+  script->abi_marker = sr.abi_marker;
   script_finalize (script, c, &sr);
 
   ia64_free_state_record (&sr);
@@ -409,14 +417,16 @@ static inline int
 run_script (struct ia64_script *script, struct cursor *c)
 {
   struct ia64_script_insn *ip, *limit, next_insn;
-  unw_word_t val, unat_addr, *s = (unw_word_t *) c;
+  unw_word_t val, unat_addr;
   unsigned long opc, dst;
+  ia64_loc_t loc;
   int ret;
 
   c->pi = script->pi;
   ip = script->insn;
   limit = script->insn + script->count;
   next_insn = *ip;
+  c->abi_marker = script->abi_marker;
 
   while (ip++ < limit)
     {
@@ -428,49 +438,52 @@ run_script (struct ia64_script *script, struct cursor *c)
       switch (opc)
 	{
 	case IA64_INSN_SET:
-	  s[dst] = val;
+	  loc = IA64_LOC_ADDR (val, 0);
 	  break;
 
-	case IA64_INSN_ADD:
-	  s[dst] += val;
+	case IA64_INSN_SET_REG:
+	  loc = IA64_LOC_REG (val, 0);
 	  break;
 
 	case IA64_INSN_ADD_PSP:
-	  s[dst] = c->psp + val;
+	  loc = IA64_LOC_ADDR (c->psp + val, 0);
 	  break;
 
 	case IA64_INSN_ADD_SP:
-	  s[dst] = c->sp + val;
+	  loc = IA64_LOC_ADDR (c->sp + val, 0);
 	  break;
 
 	case IA64_INSN_MOVE:
-	  s[dst] = s[val];
+	  loc = c->loc[val];
 	  break;
 
 	case IA64_INSN_MOVE_SCRATCH:
-	  s[dst] = ia64_scratch_loc (c, val);
+	  loc = ia64_scratch_loc (c, val);
 	  break;
 
 	case IA64_INSN_MOVE_STACKED:
 	  val = rotate_gr (c, val);
-	  ret = ia64_get_stacked (c, val, &s[dst], NULL);
-	  if (ret < 0)
-		  return ret;
+	  if ((ret = ia64_get_stacked (c, val, &loc, NULL)) < 0)
+	    return ret;
 	  break;
 
 	case IA64_INSN_SETNAT_MEMSTK:
-	  ret = ia64_get (c, c->pri_unat_loc, &unat_addr);
-	  if (ret < 0)
+	  if ((ret = ia64_get (c, c->loc[IA64_REG_PRI_UNAT_MEM],
+			       &unat_addr)) < 0)
 	    return ret;
-	  s[dst] = IA64_LOC (unat_addr >> 3, IA64_LOC_TYPE_MEMSTK_NAT);
+	  loc = IA64_LOC_ADDR (unat_addr, IA64_LOC_TYPE_MEMSTK_NAT);
 	  break;
 
-	case IA64_INSN_LOAD:
-	  ret = ia64_get (c, s[val], &s[dst]);
-	  if (ret < 0)
+	case IA64_INSN_INC_PSP:
+	  c->psp += val;
+	  continue;
+
+	case IA64_INSN_LOAD_PSP:
+	  if ((ret = ia64_get (c, c->loc[IA64_REG_PSP], &c->psp)) < 0)
 	    return ret;
-	  break;
+	  continue;
 	}
+      c->loc[dst] = loc;
     }
   return 0;
 }
@@ -478,83 +491,68 @@ run_script (struct ia64_script *script, struct cursor *c)
 HIDDEN int
 ia64_find_save_locs (struct cursor *c)
 {
-  int global_cache = (c->as->caching_policy == UNW_CACHE_GLOBAL);
   struct ia64_script_cache *cache = NULL;
   struct ia64_script *script = NULL;
   sigset_t saved_sigmask;
   int ret = 0;
 
-  cache = get_script_cache (c->as);
+  cache = get_script_cache (c->as, &saved_sigmask);
+  {
+    if (c->as->caching_policy == UNW_CACHE_NONE)
+      {
+	struct ia64_script tmp_script;
 
-  if (likely (global_cache))
-    {
-      sigprocmask (SIG_SETMASK, &unwi_full_sigmask, &saved_sigmask);
-      mutex_lock (&cache->lock);
-    }
+	if ((ret = ia64_fetch_proc_info (c, c->ip, 1)) < 0)
+	  goto out;
 
-  if (c->as->caching_policy == UNW_CACHE_NONE)
-    {
-      struct ia64_script tmp_script;
-
-      script = &tmp_script;
-      script->ip = c->ip;
-      script->hint = 0;
-      script->count = 0;
-      ret = build_script (c, script);
-    }
-  else
-    {
-      script = script_lookup (cache, c);
-      if (!script)
-	{
-	  script = script_new (cache, c->ip);
-	  if (!script)
-	    {
-	      dprintf ("%s: failed to create unwind script\n", __FUNCTION__);
-	      ret = -UNW_EUNSPEC;
+	script = &tmp_script;
+	script->ip = c->ip;
+	script->hint = 0;
+	script->count = 0;
+	ret = build_script (c, script);
+      }
+    else
+      {
+	script = script_lookup (cache, c);
+	debug (125, "%s: ip %lx %s in script cache\n",
+	       __FUNCTION__, (long) c->ip, script ? "hit" : "missed");
+	if (!script)
+	  {
+	    if ((ret = ia64_fetch_proc_info (c, c->ip, 1)) < 0)
 	      goto out;
-	    }
-	  cache->buckets[c->prev_script].hint = script - cache->buckets;
 
-	  ret = build_script (c, script);
-	}
-      c->hint = script->hint;
-      c->prev_script = script - cache->buckets;
-    }
-  if (ret < 0)
-    {
-      if (ret != UNW_ESTOPUNWIND)
-	dprintf ("%s: failed to locate/build unwind script for ip %lx\n",
-		 __FUNCTION__, (long) c->ip);
-      goto out;
-    }
+	    script = script_new (cache, c->ip);
+	    if (!script)
+	      {
+		dprintf ("%s: failed to create unwind script\n", __FUNCTION__);
+		ret = -UNW_EUNSPEC;
+		goto out;
+	      }
+	    cache->buckets[c->prev_script].hint = script - cache->buckets;
 
-  run_script (script, c);
+	    ret = build_script (c, script);
+	  }
+	c->hint = script->hint;
+	c->prev_script = script - cache->buckets;
+      }
+    if (ret < 0)
+      {
+	if (ret != -UNW_ESTOPUNWIND)
+	  dprintf ("%s: failed to locate/build unwind script for ip %lx\n",
+		   __FUNCTION__, (long) c->ip);
+	goto out;
+      }
 
+    run_script (script, c);
+  }
  out:
-  if (likely (global_cache))
-    {
-      mutex_unlock (&cache->lock);
-      sigprocmask (SIG_SETMASK, &saved_sigmask, NULL);
-    }
+  put_script_cache (c->as, cache, &saved_sigmask);
   return ret;
 }
 
 HIDDEN void
 ia64_script_cache_init (struct ia64_script_cache *cache)
 {
-  int i;
-
   mutex_init (&cache->lock);
-  cache->lru_head = IA64_UNW_CACHE_SIZE - 1;
-  cache->lru_tail = 0;
-
-  for (i = 0; i < IA64_UNW_CACHE_SIZE; ++i)
-    {
-      if (i > 0)
-	cache->buckets[i].lru_chain = (i - 1);
-      cache->buckets[i].coll_chain = -1;
-    }
-  for (i = 0; i<IA64_UNW_HASH_SIZE; ++i)
-    cache->hash[i] = -1;
+  flush_script_cache (cache);
 }
